@@ -1,33 +1,6 @@
-#!/usr/bin/env python
-# $Id: servers.py 1230 2013-07-16 16:07:08Z btimby $
-
-#  ======================================================================
-#  Copyright (C) 2007-2013 Giampaolo Rodola' <g.rodola@gmail.com>
-#
-#                         All Rights Reserved
-#
-# Permission is hereby granted, free of charge, to any person
-# obtaining a copy of this software and associated documentation
-# files (the "Software"), to deal in the Software without
-# restriction, including without limitation the rights to use,
-# copy, modify, merge, publish, distribute, sublicense, and/or sell
-# copies of the Software, and to permit persons to whom the
-# Software is furnished to do so, subject to the following
-# conditions:
-#
-# The above copyright notice and this permission notice shall be
-# included in all copies or substantial portions of the Software.
-#
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-# OTHER DEALINGS IN THE SOFTWARE.
-#
-#  ======================================================================
+# Copyright (C) 2007 Giampaolo Rodola' <g.rodola@gmail.com>.
+# Use of this source code is governed by MIT license that can be
+# found in the LICENSE file.
 
 """
 This module contains the main FTPServer class which listens on a
@@ -42,8 +15,8 @@ You might be interested in these in case your code contains blocking
 parts which cannot be adapted to the base async model or if the
 underlying filesystem is particularly slow, see:
 
-https://code.google.com/p/pyftpdlib/issues/detail?id=197
-https://code.google.com/p/pyftpdlib/issues/detail?id=212
+https://github.com/giampaolo/pyftpdlib/issues/197
+https://github.com/giampaolo/pyftpdlib/issues/212
 
 Two classes are provided:
 
@@ -60,23 +33,26 @@ This way the handler handling that connections will be free to block
 without hanging the whole FTP server.
 """
 
-import os
-import socket
-import traceback
-import sys
 import errno
+import os
 import select
-import logging
 import signal
+import sys
+import threading
 import time
+import traceback
 
-from pyftpdlib.log import logger
-from pyftpdlib.ioloop import Acceptor, IOLoop
+from .ioloop import Acceptor
+from .ioloop import IOLoop
+from .log import config_logging
+from .log import debug
+from .log import is_logging_configured
+from .log import logger
 
-
-__all__ = ['FTPServer']
+__all__ = ['FTPServer', 'ThreadedFTPServer']
 _BSD = 'bsd' in sys.platform
 
+
 # ===================================================================
 # --- base class
 # ===================================================================
@@ -105,7 +81,7 @@ class FTPServer(Acceptor):
     max_cons = 512
     max_cons_per_ip = 0
 
-    def __init__(self, address_or_socket, handler, ioloop=None, backlog=5):
+    def __init__(self, address_or_socket, handler, ioloop=None, backlog=100):
         """Creates a socket listening on 'address' dispatching
         connections to a 'handler'.
 
@@ -134,17 +110,16 @@ class FTPServer(Acceptor):
             sock = address_or_socket
             sock.setblocking(0)
             self.set_socket(sock)
-            if hasattr(sock, 'family'):
-                self._af = sock.family
-            else:
-                # python 2.4
-                ip, port = self.socket.getsockname()[:2]
-                self._af = socket.getaddrinfo(ip, port, socket.AF_UNSPEC,
-                                              socket.SOCK_STREAM)[0][0]
         else:
-            self._af = self.bind_af_unspecified(address_or_socket)
+            self.bind_af_unspecified(address_or_socket)
         self.listen(backlog)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *args):
+        self.close_all()
+
     @property
     def address(self):
         return self.socket.getsockname()[:2]
@@ -160,13 +135,20 @@ class FTPServer(Acceptor):
             return self._map_len() <= self.max_cons
 
     def _log_start(self):
-        if not logging.getLogger().handlers:
+        def get_fqname(obj):
+            try:
+                return obj.__module__ + "." + obj.__class__.__name__
+            except AttributeError:
+                try:
+                    return obj.__module__ + "." + obj.__name__
+                except AttributeError:
+                    return str(obj)
+
+        if not is_logging_configured():
             # If we get to this point it means the user hasn't
-            # configured logger. We want to log by default so
-            # we configure logging ourselves so that it will
-            # print to stderr.
-            from pyftpdlib.ioloop import _config_logging
-            _config_logging()
+            # configured any logger. We want logging to be on
+            # by default (stderr).
+            config_logging()
 
         if self.handler.passive_ports:
             pasv_ports = "%s->%s" % (self.handler.passive_ports[0],
@@ -174,20 +156,45 @@ class FTPServer(Acceptor):
         else:
             pasv_ports = None
         addr = self.address
-        logger.info(">>> starting FTP server on %s:%s, pid=%i <<<"
-                    % (addr[0], addr[1], os.getpid()))
-        logger.info("poller: %r", self.ioloop.__class__)
+        if hasattr(self.handler, 'ssl_protocol'):
+            proto = "FTP+SSL"
+        else:
+            proto = "FTP"
+        logger.info(">>> starting %s server on %s:%s, pid=%i <<<"
+                    % (proto, addr[0], addr[1], os.getpid()))
+        if ('ThreadedFTPServer' in __all__ and
+                issubclass(self.__class__, ThreadedFTPServer)):
+            logger.info("concurrency model: multi-thread")
+        elif ('MultiprocessFTPServer' in __all__ and
+                issubclass(self.__class__, MultiprocessFTPServer)):
+            logger.info("concurrency model: multi-process")
+        elif issubclass(self.__class__, FTPServer):
+            logger.info("concurrency model: async")
+
         logger.info("masquerade (NAT) address: %s",
                     self.handler.masquerade_address)
         logger.info("passive ports: %s", pasv_ports)
+        logger.debug("poller: %r", get_fqname(self.ioloop))
+        logger.debug("authorizer: %r", get_fqname(self.handler.authorizer))
         if os.name == 'posix':
-            logger.info("use sendfile(2): %s", self.handler.use_sendfile)
+            logger.debug("use sendfile(2): %s", self.handler.use_sendfile)
+        logger.debug("handler: %r", get_fqname(self.handler))
+        logger.debug("max connections: %s", self.max_cons or "unlimited")
+        logger.debug("max connections per ip: %s",
+                     self.max_cons_per_ip or "unlimited")
+        logger.debug("timeout: %s", self.handler.timeout or "unlimited")
+        logger.debug("banner: %r", self.handler.banner)
+        logger.debug("max login attempts: %r", self.handler.max_login_attempts)
+        if getattr(self.handler, 'certfile', None):
+            logger.debug("SSL certfile: %r", self.handler.certfile)
+        if getattr(self.handler, 'keyfile', None):
+            logger.debug("SSL keyfile: %r", self.handler.keyfile)
 
     def serve_forever(self, timeout=None, blocking=True, handle_exit=True):
         """Start serving.
 
          - (float) timeout: the timeout passed to the underlying IO
-           loop expressed in seconds (default 1.0).
+           loop expressed in seconds.
 
          - (bool) blocking: if False loop once and then return the
            timeout of the next scheduled call next to expire soonest
@@ -199,17 +206,19 @@ class FTPServer(Acceptor):
            Also, logs server start and stop.
         """
         if handle_exit:
-            log = handle_exit and blocking == True
+            log = handle_exit and blocking
             if log:
                 self._log_start()
             try:
                 self.ioloop.loop(timeout, blocking)
             except (KeyboardInterrupt, SystemExit):
-                pass
+                logger.info("received interrupt signal")
             if blocking:
                 if log:
-                    logger.info(">>> shutting down FTP server (%s active fds) <<<",
-                                self._map_len())
+                    logger.info(
+                        ">>> shutting down FTP server (%s active socket "
+                        "fds) <<<",
+                        self._map_len())
                 self.close_all()
         else:
             self.ioloop.loop(timeout, blocking)
@@ -244,7 +253,7 @@ class FTPServer(Acceptor):
 
             try:
                 handler.handle()
-            except:
+            except Exception:
                 handler.handle_error()
             else:
                 return handler
@@ -253,8 +262,8 @@ class FTPServer(Acceptor):
             # be fixed. We do not want to tear down the server though
             # (DoS). We just log the exception, hoping that someone
             # will eventually file a bug. References:
-            # - http://code.google.com/p/pyftpdlib/issues/detail?id=143
-            # - http://code.google.com/p/pyftpdlib/issues/detail?id=166
+            # - https://github.com/giampaolo/pyftpdlib/issues/143
+            # - https://github.com/giampaolo/pyftpdlib/issues/166
             # - https://groups.google.com/forum/#!topic/pyftpdlib/h7pPybzAx14
             logger.error(traceback.format_exc())
             if handler is not None:
@@ -287,37 +296,64 @@ class _SpawnerBase(FTPServer):
     Not supposed to be used.
     """
 
-    # how many seconds to wait when join()ing parent's threads
-    # or processes
+    # How many seconds to wait when join()ing parent's threads
+    # or processes.
     join_timeout = 5
+    # How often thread/process finished tasks should be cleaned up.
+    refresh_interval = 5
     _lock = None
     _exit = None
 
-    def __init__(self, address, handler, ioloop=None):
-        FTPServer.__init__(self, address, handler, ioloop)
+    def __init__(self, address_or_socket, handler, ioloop=None, backlog=100):
+        FTPServer.__init__(self, address_or_socket, handler,
+                           ioloop=ioloop, backlog=backlog)
         self._active_tasks = []
+        self._active_tasks_idler = self.ioloop.call_every(
+            self.refresh_interval,
+            self._refresh_tasks,
+            _errback=self.handle_error)
 
     def _start_task(self, *args, **kwargs):
         raise NotImplementedError('must be implemented in subclass')
 
-    def _current_task(self):
-        raise NotImplementedError('must be implemented in subclass')
-
     def _map_len(self):
-        raise NotImplementedError('must be implemented in subclass')
+        if len(self._active_tasks) >= self.max_cons:
+            # Since refresh()ing is a potentially expensive operation
+            # (O(N)) do it only if we're exceeding max connections
+            # limit. Other than in here, tasks are refreshed every 10
+            # seconds anyway.
+            self._refresh_tasks()
+        return len(self._active_tasks)
+
+    def _refresh_tasks(self):
+        """join() terminated tasks and update internal _tasks list.
+        This gets called every X secs.
+        """
+        if self._active_tasks:
+            logger.debug("refreshing tasks (%s join() potentials)" %
+                         len(self._active_tasks))
+            with self._lock:
+                new = []
+                for t in self._active_tasks:
+                    if not t.is_alive():
+                        self._join_task(t)
+                    else:
+                        new.append(t)
+
+                self._active_tasks = new
 
     def _loop(self, handler):
         """Serve handler's IO loop in a separate thread or process."""
-        ioloop = IOLoop()
-        try:
+        with IOLoop() as ioloop:
             handler.ioloop = ioloop
             try:
                 handler.add_channel()
-            except EnvironmentError:
-                err = sys.exc_info()[1]
+            except EnvironmentError as err:
                 if err.errno == errno.EBADF:
                     # we might get here in case the other end quickly
                     # disconnected (see test_quick_connect())
+                    debug("call: %s._loop(); add_channel() returned EBADF",
+                          self)
                     return
                 else:
                     raise
@@ -328,8 +364,8 @@ class _SpawnerBase(FTPServer):
             poll_timeout = getattr(self, 'poll_timeout', None)
             soonest_timeout = poll_timeout
 
-            while (ioloop.socket_map or ioloop.sched._tasks) and not \
-              self._exit.is_set():
+            while (ioloop.socket_map or ioloop.sched._tasks) and \
+                    not self._exit.is_set():
                 try:
                     if ioloop.socket_map:
                         poll(timeout=soonest_timeout)
@@ -342,9 +378,10 @@ class _SpawnerBase(FTPServer):
                         # functions are supposed to be cancel()ed on close()
                         # but by using threads we can incur into
                         # synchronization issues such as this one.
-                        # https://code.google.com/p/pyftpdlib/issues/detail?id=245
+                        # https://github.com/giampaolo/pyftpdlib/issues/245
                         if not ioloop.socket_map:
-                            ioloop.sched.reheapify() # get rid of cancel()led calls
+                            # get rid of cancel()led calls
+                            ioloop.sched.reheapify()
                             soonest_timeout = sched_poll()
                             if soonest_timeout:
                                 time.sleep(min(soonest_timeout, 1))
@@ -354,11 +391,10 @@ class _SpawnerBase(FTPServer):
                     # note: these two exceptions are raised in all sub
                     # processes
                     self._exit.set()
-                except select.error:
+                except select.error as err:
                     # on Windows we can get WSAENOTSOCK if the client
                     # rapidly connect and disconnects
-                    err = sys.exc_info()[1]
-                    if os.name == 'nt' and err.args[0] == 10038:
+                    if os.name == 'nt' and err[0] == 10038:
                         for fd in list(ioloop.socket_map.keys()):
                             try:
                                 select.select([fd], [], [], 0)
@@ -374,11 +410,9 @@ class _SpawnerBase(FTPServer):
                         raise
                 else:
                     if poll_timeout:
-                        if soonest_timeout is None \
-                        or soonest_timeout > poll_timeout:
+                        if (soonest_timeout is None or
+                                soonest_timeout > poll_timeout):
                             soonest_timeout = poll_timeout
-        finally:
-            ioloop.close()
 
     def handle_accepted(self, sock, addr):
         handler = FTPServer.handle_accepted(self, sock, addr)
@@ -387,7 +421,8 @@ class _SpawnerBase(FTPServer):
             # main thread to accept connections
             self.ioloop.unregister(handler._fileno)
 
-            t = self._start_task(target=self._loop, args=(handler,))
+            t = self._start_task(target=self._loop, args=(handler, ),
+                                 name='ftpd')
             t.name = repr(addr)
             t.start()
 
@@ -395,25 +430,17 @@ class _SpawnerBase(FTPServer):
             if hasattr(t, 'pid'):
                 handler.close()
 
-            self._lock.acquire()
-            try:
-                # clean finished tasks
-                for task in self._active_tasks[:]:
-                    if not task.is_alive():
-                        self._active_tasks.remove(task)
+            with self._lock:
                 # add the new task
                 self._active_tasks.append(t)
-            finally:
-                self._lock.release()
 
     def _log_start(self):
         FTPServer._log_start(self)
-        logger.info("dispatcher: %r", self.__class__)
 
-    def serve_forever(self, timeout=None, blocking=True, handle_exit=True):
+    def serve_forever(self, timeout=1.0, blocking=True, handle_exit=True):
         self._exit.clear()
         if handle_exit:
-            log = handle_exit and blocking == True
+            log = handle_exit and blocking
             if log:
                 self._log_start()
             try:
@@ -422,119 +449,80 @@ class _SpawnerBase(FTPServer):
                 pass
             if blocking:
                 if log:
-                    logger.info(">>> shutting down FTP server (%s active " \
-                                "workers) <<<", self._map_len())
+                    logger.info(
+                        ">>> shutting down FTP server (%s active workers) <<<",
+                        self._map_len())
                 self.close_all()
         else:
             self.ioloop.loop(timeout, blocking)
 
+    def _terminate_task(self, t):
+        if hasattr(t, 'terminate'):
+            logger.debug("terminate()ing task %r" % t)
+            try:
+                if not _BSD:
+                    t.terminate()
+                else:
+                    # XXX - On FreeBSD using SIGTERM doesn't work
+                    # as the process hangs on kqueue.control() or
+                    # select.select(). Use SIGKILL instead.
+                    os.kill(t.pid, signal.SIGKILL)
+            except OSError as err:
+                if err.errno != errno.ESRCH:
+                    raise
+
+    def _join_task(self, t):
+        logger.debug("join()ing task %r" % t)
+        t.join(self.join_timeout)
+        if t.is_alive():
+            logger.warning("task %r remained alive after %r secs", t,
+                           self.join_timeout)
+
     def close_all(self):
-        tasks = self._active_tasks[:]
+        self._active_tasks_idler.cancel()
         # this must be set after getting active tasks as it causes
         # thread objects to get out of the list too soon
         self._exit.set()
-        if tasks and hasattr(tasks[0], 'terminate'):
-            # we're dealing with subprocesses
-            for t in tasks:
-                try:
-                    if not _BSD:
-                        t.terminate()
-                    else:
-                        # XXX - On FreeBSD using SIGTERM doesn't work
-                        # as the process hangs on kqueue.control() or
-                        # select.select(). Use SIGKILL instead.
-                        os.kill(t.pid, signal.SIGKILL)
-                except OSError:
-                    err = sys.exc_info()[1]
-                    if err.errno != errno.ESRCH:
-                        raise
 
-        self._wait_for_tasks(tasks)
-        del self._active_tasks[:]
+        with self._lock:
+            for t in self._active_tasks:
+                self._terminate_task(t)
+            for t in self._active_tasks:
+                self._join_task(t)
+            del self._active_tasks[:]
+
         FTPServer.close_all(self)
 
-    def _wait_for_tasks(self, tasks):
-        """Wait for threads or subprocesses to terminate."""
-        warn = logger.warning
-        for t in tasks:
-            t.join(self.join_timeout)
-            if t.is_alive():
-                # Thread or process is still alive. If it's a process
-                # attempt to send SIGKILL as last resort.
-                # Set timeout to None so that we will exit immediately
-                # in case also other threads/processes are hanging.
-                self.join_timeout = None
-                if hasattr(t, 'terminate'):
-                    msg = "could not terminate process %r" % t
-                    if not _BSD:
-                        warn(msg + "; sending SIGKILL as last resort")
-                        try:
-                            os.kill(t.pid, signal.SIGKILL)
-                        except OSError:
-                            err = sys.exc_info()[1]
-                            if err.errno != errno.ESRCH:
-                                raise
-                    else:
-                        warn(msg)
-                else:
-                    warn("thread %r didn't terminate; ignoring it", t)
 
-
-try:
-    import threading
-except ImportError:
-    pass
-else:
-    __all__ += ['ThreadedFTPServer']
+class ThreadedFTPServer(_SpawnerBase):
+    """A modified version of base FTPServer class which spawns a
+    thread every time a new connection is established.
+    """
+    # The timeout passed to thread's IOLoop.poll() call on every
+    # loop. Necessary since threads ignore KeyboardInterrupt.
+    poll_timeout = 1.0
+    _lock = threading.Lock()
+    _exit = threading.Event()
 
     # compatibility with python <= 2.6
-    if not hasattr(threading.Thread, 'is_alive'):
-        threading.Thread.is_alive = threading.Thread.isAlive
+    if not hasattr(_exit, 'is_set'):
+        _exit.is_set = _exit.isSet
 
-    class ThreadedFTPServer(_SpawnerBase):
-        """A modified version of base FTPServer class which spawns a
-        thread every time a new connection is established.
-        """
-        # The timeout passed to thread's IOLoop.poll() call on every
-        # loop. Necessary since threads ignore KeyboardInterrupt.
-        poll_timeout = 1.0
-        _lock = threading.Lock()
-        _exit = threading.Event()
-
-        # compatibility with python <= 2.6
-        if not hasattr(_exit, 'is_set'):
-            _exit.is_set = _exit.isSet
-
-        def _start_task(self, *args, **kwargs):
-            return threading.Thread(*args, **kwargs)
-
-        def _current_task(self):
-            return threading.currentThread()
-
-        def _map_len(self):
-            return threading.activeCount()
+    def _start_task(self, *args, **kwargs):
+        return threading.Thread(*args, **kwargs)
 
 
 if os.name == 'posix':
-    try:
-        import multiprocessing
-    except ImportError:
-        pass
-    else:
-        __all__ += ['MultiprocessFTPServer']
+    import multiprocessing
 
-        class MultiprocessFTPServer(_SpawnerBase):
-            """A modified version of base FTPServer class which spawns a
-            process every time a new connection is established.
-            """
-            _lock = multiprocessing.Lock()
-            _exit = multiprocessing.Event()
+    __all__ += ['MultiprocessFTPServer']
 
-            def _start_task(self, *args, **kwargs):
-                return multiprocessing.Process(*args, **kwargs)
+    class MultiprocessFTPServer(_SpawnerBase):
+        """A modified version of base FTPServer class which spawns a
+        process every time a new connection is established.
+        """
+        _lock = multiprocessing.Lock()
+        _exit = multiprocessing.Event()
 
-            def _current_task(self):
-                return multiprocessing.current_process()
-
-            def _map_len(self):
-                return len(multiprocessing.active_children())
+        def _start_task(self, *args, **kwargs):
+            return multiprocessing.Process(*args, **kwargs)
